
class CriticalErrorException(Exception):
    """Exception raised for critical errors in the cobra_log module."""


class ErrorException(Exception):
    """Exception raised for errors in the cobra_log module."""


class WarningException(Exception):
    """Exception raised for warnings in the cobra_log module."""
